/******************************************************************************
 * IP protocols logger					                      *
 ******************************************************************************
 * Written by Hugo Haas (hugo@debian.org)                                     *
 * Based on the idea of Mike Edulla (medulla@infosoc.com)                     *
 * Copyright (C) 1998-1999 Hugo Haas                                          *
 * See the HISTORY file for the changes					      *
 *                                                                            *
 * Use of getopt_long for argument parsing: Etienne Bernard 1999              *
 ******************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <stdlib.h>
#include <getopt.h>
#include <signal.h>

#include "defines.h"
#include "configuration.h"
#include "icmp.h"
#include "tcp.h"
#include "udp.h"
#include "log.h"
#include "netutils.h"
#include "filter.h"
#include "pidfile.h"

/* Do we have to run as a daemon? */
int run_as_daemon = TRUE;

/* Should name resolving be done? */
int dns_resolve = TRUE;

/* Should we log more details? */
int log_details = FALSE;

/* Configuration file to use */
char *configuration_file = NULL;

/* Logging mechanism */
struct logger log = {
#if 0
  NULL,
#endif
  NULL };

/* Protocols that we log */
unsigned short log_protocols;

/* Expiration of DNS data */
unsigned int dns_expire = DNS_EXPIRE;

/* Threads */
pthread_t icmp_t, tcp_t, udp_t;

/*
 * do_help
 *
 * Display the command line options
 */

void do_help() {

  printf("ippl version %s \n\n", VERSION);
  printf("\t-n, --nodaemon:\tdo not run as a daemon and write information to stdout\n");
  printf("\t-r, --noresolve:\tno resolving of IP addresses \n");
  printf("\t-l, --long:\tgive more detailed information \n");
  printf("\t-c, --config <file>:\tspecify an alternate configuration file\n");
  printf("\t-h, --help:\tdisplay this help and exit\n");
  printf("\nPlease report bugs to ippl@via.ecp.fr\n");
}

/*
 * run_thread
 *
 * Run a thread. Certain signals will are blocked.
 *
 */

void run_thread(pthread_t *t, void *function, void *param) {
  pthread_attr_t attr_t;
  sigset_t newset, oldset ;

  sigfillset(&newset);
  pthread_sigmask(SIG_BLOCK, &newset, &oldset);

  pthread_attr_init(&attr_t);
  pthread_attr_setdetachstate(&attr_t, PTHREAD_CREATE_DETACHED);
  pthread_create(t, &attr_t, function, param);
  pthread_attr_destroy(&attr_t);

  pthread_sigmask(SIG_UNBLOCK, &newset, &oldset);
}


/*
 * start_all_threads
 *
 * Parses the config file and launches the threads
 */
void start_all_threads() {
  struct passwd *nobody;

  nobody = getpwnam("nobody");
  if (!nobody) {
    printf("I cannot find the \"nobody\" account.\n");
    exit(1);
  }

  parse_config_file(configuration_file);
  
  if (log_protocols == NONE) {
    printf("Nothing to log.\n");
    exit(2);
  }
      
  initialize_dns_cache();
      
  if (log_protocols & RUN_ICMP)
    run_thread(&icmp_t, (void *) log_icmp, (void *)nobody);
      
  if (log_protocols & RUN_TCP)
    run_thread(&tcp_t, (void *) log_tcp, (void *)nobody);
      
  if (log_protocols & RUN_UDP)
    run_thread(&udp_t, (void *) log_udp, (void *)nobody);
}  


/*
 * reload_configuration
 *
 * Stops the threads and reloads the configuration
 */
void reload_configuration() {
  struct filter_entry *p;
  extern pthread_mutex_t mutex, service_mutex, dns_mutex, r_mux, w_mux;
  extern pthread_cond_t w_cond;
  extern struct filter_entry * filter;
  extern int readers;

  /* We get all the mutexes that could block the threads,
   * because we must be sure that no thread is in a
   * critical portion of the code. */

  /* First, make sure that no packet is being logged
   * Let's block acces to the filter */
  pthread_mutex_lock(&w_mux);
  pthread_mutex_lock(&r_mux);
  while (readers > 0)
    pthread_cond_wait(&w_cond, &r_mux);
  /* Then, acquire all the others so that you are sure we are in a safe
   * position */
  /* dns cache mutex */
  pthread_mutex_lock(&dns_mutex);
  /* Service query mutex */
  pthread_mutex_lock(&service_mutex);
  /* Logging mutex */
  pthread_mutex_lock(&mutex);

  /* Now we can safely kill the threads. */
  if (log_protocols & RUN_ICMP)
    pthread_cancel(icmp_t);    
  if (log_protocols & RUN_TCP)
    pthread_cancel(tcp_t);
  if (log_protocols & RUN_UDP)
    pthread_cancel(udp_t);

  /* We release the mutexes as soon as possible */
  pthread_mutex_unlock(&mutex);
  pthread_mutex_unlock(&service_mutex);
  pthread_mutex_unlock(&dns_mutex);
  pthread_mutex_unlock(&r_mux);
  pthread_mutex_unlock(&w_mux);

  /* We destroy previous configuration */
  while ((p = filter) != NULL) {
    filter = filter->next;
    free(p);
  }

  /* And we restart all the threads */
  start_all_threads();
}

/*
 * parse_arguments
 *
 * Parse the arguments given on the command line
 */

void parse_arguments(int argc, char **argv) {
  static struct option long_options[] =
  {
    { "help", 0, 0, 0 },
    { "nodaemon", 0, &run_as_daemon, FALSE },
    { "noresolve", 0, &dns_resolve, FALSE },
    { "long", 0, &log_details, TRUE },
    { "config", 1, 0, 0 },
    { 0, 0, 0, 0 }
  };

  int c, option_index = 0;

  while ((c = getopt_long(argc, argv, "hnlrc:", long_options,
                          &option_index))!=-1) {
    switch (c) {
    case 0:
      switch (option_index) {
      case 0:
        do_help();
        exit(0);
      case 2:
        configuration_file = optarg;
        break;
      }
      break;
    case 'h':
      do_help();
      exit(0);
    case 'n':
      run_as_daemon = FALSE;
      break;
    case 'l':
      log_details = TRUE;
      break;
    case 'r':
      dns_resolve = FALSE;
      break;
    case 'c':
      configuration_file = optarg;
      break;
    case '?':
      do_help();
      exit(1);
    }
  }
}

/*
 * go_background
 *
 * Run as a daemon
 *
 * Based on iplogger (C) Mike Edulla + modifications by Topi Miettinen
 * PID file handling based on sysklogd (C) Martin Schulze
 */

void go_background() {
  int     fd;
  int     fs;

  /* Check PID */
  if (check_pid(PID_FILE)) {
    log.f(LOGLEVEL, "Already running. Exiting...");
    exit(1);
  }

  if (getppid() != 1) {
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    fs = fork();
    if (fs < 0) {
      perror("fork");
      exit(1);
    }
    if (fs > 0) {
      exit(0);
    }
    fd = open("/dev/tty", O_RDWR);
    if (fd >= 0) {
      ioctl(fd, TIOCNOTTY, (char *) NULL);
      close(fd);
    }
  }
  for (fd = 0; fd < NOFILE; fd++)
    close(fd);
  errno = 0;
  umask(0);
  setsid();

  /* Write PID file */
  if (!write_pid(PID_FILE)) {
    log.f(LOGLEVEL,"Can't write pid.\n");
    exit(1);
  }

}

/*
 * die
 *
 * Function executed when the daemon is ended
 */

void die(int sig) {

  if (run_as_daemon) {
    (void) remove_pid(PID_FILE);
  }

  log.f(LOGLEVEL, "IP Protocols Logger stopped (signal %d).", sig);

  exit(0);
}

/*
 * sighup
 *
 * Function executed when we receive a SIHUP signal
 */
void sighup(int sig) {
  reload_configuration();
  log.f(LOGLEVEL, "reloaded configuration.");
  signal(SIGHUP, sighup);
}

/*
 * handle_signals
 *
 * Define how to handle the different signals
 */

void handle_signals(void) {

  signal(SIGTERM, die);
  if (!run_as_daemon) {
    signal(SIGINT, die);
  }
  signal(SIGHUP, sighup);
}

/*
 * MAIN
 */

int main(int argc, char **argv) {
  struct passwd *nobody;

  parse_arguments(argc, argv);

  setuid(0);
  if (geteuid() != 0) {
    printf("You must be root to run this program.\n");
    exit(1);
  }

  nobody = getpwnam("nobody");
  if (!nobody) {
    printf("I cannot find the \"nobody\" account.\n");
    exit(1);
  }

  handle_signals();

  if (configuration_file == NULL)
    configuration_file = strdup(CONFIGURATION_FILE);

  if (run_as_daemon) {
#if 0
    if (log.file)
      log.f = (void *) filelog;
    else
#endif
    log.f = (void *) syslog;
    go_background();
  }
  else {
    log.f = (void *) stdoutlog;
  }
      
  start_all_threads();
  log.f(LOGLEVEL, "IP Protocols Logger started.");

  for(;;) {
    sleep(dns_expire);
    initialize_dns_cache();
    refresh_filter();
  }

  free(nobody);

  exit(0);
}
